/*
This file is part of myFitter, a C++ library for maximum likelihood fits.
Copyright (C) 2012  Martin Wiebusch

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef MYFITTER_LIKELIHOODCOMPONENTS_HPP
#define MYFITTER_LIKELIHOODCOMPONENTS_HPP

#include "likelihoodfunction.hpp"


namespace myfitter {

// single observables with gaussian and systematic errors.

class GaussianLC : public LikelihoodComponent {
private:
    double stat_err_;
    double syst_minus_;
    double syst_plus_;

public:
    // constructors
    GaussianLC(int iobs, double mean, double stat_err,
               double syst_plus=0., double syst_minus=0.) :
        LikelihoodComponent(iobs+1), stat_err_(stat_err),
        syst_minus_(syst_minus), syst_plus_(syst_plus) {
        add_(iobs, mean);
    }
    GaussianLC(const GaussianLC& lc) :
        LikelihoodComponent(lc), stat_err_(lc.stat_err_),
        syst_minus_(lc.syst_minus_), syst_plus_(lc.syst_plus_) { }
    virtual GaussianLC* copy() const
    { return new GaussianLC(*this); }
    
    virtual double calc(const ObservableVector& pred,
                        const ObservableVector& meas);

    virtual int calc_deriv(const ObservableVector& pred,
                           const ObservableVector& meas,
                           ObservableVector& result);

    virtual double calc_nosystematics(const ObservableVector& pred,
                                      const ObservableVector& meas);

    virtual void get_hessian(Matrix& hess, const ObservableVector& obs) {
        hess(claimed_observables()[0], claimed_observables()[0]) =
            1./(stat_err_*stat_err_);
    }
};


// asymmetric Gaussian

class AsymmetricGaussianLC : public LikelihoodComponent {
private:
    double stat_minus_;
    double stat_plus_;
    double syst_minus_;
    double syst_plus_;

public:
    // constructors
    AsymmetricGaussianLC(int iobs, double mean,
                         double stat_plus, double stat_minus,
                         double syst_plus=0., double syst_minus=0.) :
        LikelihoodComponent(iobs+1),
        stat_minus_(stat_minus), stat_plus_(stat_plus),
        syst_minus_(syst_minus), syst_plus_(syst_plus)
    { add_(iobs, mean); }
    AsymmetricGaussianLC(const AsymmetricGaussianLC& lc) :
        LikelihoodComponent(lc),
        stat_minus_(lc.stat_minus_), stat_plus_(lc.stat_plus_),
        syst_minus_(lc.syst_minus_), syst_plus_(lc.syst_plus_) { }
    virtual AsymmetricGaussianLC* copy() const
    { return new AsymmetricGaussianLC(*this); }
    
    virtual double calc(const ObservableVector& pred,
                        const ObservableVector& meas);

    virtual int calc_deriv(const ObservableVector& pred,
                           const ObservableVector& meas,
                           ObservableVector& result);

    virtual double calc_nosystematics(const ObservableVector& pred,
                                      const ObservableVector& meas);

    virtual void get_hessian(Matrix& hess, const ObservableVector& obs) {
        hess(claimed_observables()[0], claimed_observables()[0]) =
            1./std::pow(0.5*(stat_plus_ + stat_minus_), 2);
    }
};


// correlated observables with gaussian errors

class CorrelatedGaussianLC : public LikelihoodComponent {
private:
    ObservableVector err_;
    boost::numeric::ublas::matrix<double> cor_, invvar_;
    bool needs_update_;

    void update_();

public:
    CorrelatedGaussianLC(int nobs) 
        : LikelihoodComponent(nobs), needs_update_(true) { }
    CorrelatedGaussianLC(const CorrelatedGaussianLC& c) :
        LikelihoodComponent(c), err_(c.err_), cor_(c.cor_),
        invvar_(c.invvar_), needs_update_(c.needs_update_) { }
    virtual CorrelatedGaussianLC* copy() const
    { return new CorrelatedGaussianLC(*this); }

    // claim observable with index iobs
    void add(int iobs, double val, double err);
    void add(int iobs, double val) { add(iobs, val, Model::nan()); }
    void add(int iobs) { add(iobs, Model::nan(), Model::nan()); }
    
    // get/set central value of observable with index iobs
    double central_value(int iobs) const
    { return central_value_(internal_index_(iobs)); }
    void central_value(int iobs, double x) 
    { central_value_(internal_index_(iobs), x); }

    // get/set standard deviation of observable with index iobs
    double sigma(int iobs) const
    { return err_[internal_index_(iobs)]; }
    void sigma(int iobs, double x)
    { err_[internal_index_(iobs)] = x; needs_update_ = true; }

    // get/set correlation of observables i and j
    double cor(int i, int j) const {
        return cor_(internal_index_(i), internal_index_(j));
    }
    void cor(int i, int j, double x) {
        if(i != j) {
            cor_(internal_index_(i), internal_index_(j))
                = cor_(internal_index_(j), internal_index_(i)) = x;
            needs_update_ = true;
        }
    }

    virtual double calc(const ObservableVector& pred,
                        const ObservableVector& meas);

    virtual int calc_deriv(const ObservableVector& pred,
                              const ObservableVector& meas,
                              ObservableVector& result);

    virtual void get_hessian(Matrix& hess, const ObservableVector& obs);
};

}

#endif
