/*
This file is part of myFitter, a C++ library for maximum likelihood fits.
Copyright (C) 2012  Martin Wiebusch

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef MYFITTER_LIKELIHOODFUNCTION_HPP
#define MYFITTER_LIKELIHOODFUNCTION_HPP

#include "model.hpp"


namespace myfitter {

typedef std::vector<int> IndexVector;


// base class for likelihood-components

class LikelihoodComponent {
private:
    int nobservables_;
    IndexVector claimed_observables_;
    IndexVector obs_pos_;
    ObservableVector central_values_;

protected:
    // vector of central values. Uses internal indices.

    // claim observable with external index i
    void add_(int iobs, double val) {
        assert(iobs < nobservables());
        assert(iobs >= 0);
        obs_pos_[iobs] = n_claimed_observables();
        claimed_observables_.push_back(iobs);
        central_values_.resize(claimed_observables_.size());
        central_values_[claimed_observables_.size()-1] = val;
    }

    // convert internal to external index
    int external_index_(int i) const { return claimed_observables_[i]; }

    // convert external to internal index
    int internal_index_(int i) const 
    { assert(obs_pos_[i] >= 0); return obs_pos_[i]; }

    // access central values with internal indices
    double central_value_(int i) const { return central_values_[i]; }
    void central_value_(int i, double val) { central_values_[i] = val; }

public:
    // constructors
    LikelihoodComponent(int nobs) : nobservables_(nobs), obs_pos_(nobs, -1) { }
    LikelihoodComponent(const LikelihoodComponent& lc) :
        nobservables_(lc.nobservables()),
        claimed_observables_(lc.claimed_observables()),
        obs_pos_(lc.obs_pos_), central_values_(lc.central_values_) { }

    // virtual copy constructor
    // DERIVED CLASSES MUST OVERLOAD THIS!!
    virtual LikelihoodComponent* copy() const { return 0; }

    // number of observables required by the likelihood-component.
    // The number of observables in the likelihood-function may be bigger.
    int nobservables() const { return nobservables_; }

    // returns the vector of indices of claimed observables.
    // The claimed observables enter this component of the likelihood-fucntion
    // and may not be claimed by other components of the same
    // likelihood-function.
    const IndexVector& claimed_observables() const
    { return claimed_observables_; }

    // returns the number of claimed observables
    int n_claimed_observables() const { return claimed_observables_.size(); }

    // sets the elements of obs to the default central values of the
    // observables. Only elements corresponding to claimed observables are
    // changed.
    void get_central_values(ObservableVector& obs);

    // calculates the likelihood-component.
    // Derived classes must implement this.
    virtual double calc(const ObservableVector& pred,
                        const ObservableVector& meas) = 0;

    // calculates the derivatives with respect to the predicted observables
    // (first argument to calc()). Only those entries of result corresponding
    // to claimed observables should be set. Derived classes must implement
    // this.
    virtual int calc_deriv(const ObservableVector& pred,
                           const ObservableVector& meas,
                           ObservableVector& result) = 0;

    // calculates the likelihood-component without systematic errors.
    // Likelihood components which implement systematics should overload this
    // function.
    virtual double calc_nosystematics(const ObservableVector& pred,
                                      const ObservableVector& meas)
    { return calc(pred, meas); }

    // fill the entries of the matrix hess corresponding to (pairs of) claimed
    // observables with the elements of the Hessian of the likelihood function
    // at pred=meas=obs. The Hessian is the quadratic term in the Taylor
    // expansion of calc_nosystematics(obs, obs), considered as a function of
    // obs (without the factor 1/2).
    virtual void get_hessian(Matrix& hess,
                             const ObservableVector& obs) = 0;
};


// class representing a likelihood function.
// Serves as container for likelihood-components.

class LikelihoodFunction {
private:
    int nobs_;
    typedef std::vector<LikelihoodComponent*> LikelihoodComponentVector;
    LikelihoodComponentVector lcv_;
    IndexVector claimed_observables_;

public:
    // constructor
    LikelihoodFunction(int nobs) : nobs_(nobs) { }

    // copy constructor
    LikelihoodFunction(const LikelihoodFunction& f) :
        nobs_(f.nobs_), lcv_(f.lcv_.size()),
        claimed_observables_(f.claimed_observables_) {
        for(int i=0; i < int(f.lcv_.size()); ++i)
            lcv_[i] = f.lcv_[i]->copy();
    }

    // destructor
    ~LikelihoodFunction() {
        for(LikelihoodComponentVector::const_iterator ilcv
                = lcv_.begin(); ilcv != lcv_.end(); ++ilcv)
            delete *ilcv;
    }

    // number of observables considered.
    // Not all of these observables have to be claimed by likelihood-components.
    // Accordingly, not all observables must enter the likelihood-function.
    int nobservables() const { return nobs_; }

    // add likelihood component
    int add(const LikelihoodComponent& lc);

    // get default central values from likelihood components
    void get_central_values(ObservableVector& meas);

    // calculate the log-likelihood
    double calc(const ObservableVector& pred,
                const ObservableVector& meas);

    // calculates the derivatives of the log likelihood function with respect
    // to the predicted observables (first argument of calc()).
    int calc_deriv(const ObservableVector& pred,
                   const ObservableVector& meas,
                   ObservableVector& result);

    // calculate the log-likelihood without systematic errors
    double calc_nosystematics(const ObservableVector& pred,
                              const ObservableVector& meas);

    // returns the vector of indices of claimed observables.
    const IndexVector& claimed_observables() { return claimed_observables_; }

    // returns the number of claimed observables
    int n_claimed_observables() const { return claimed_observables_.size(); }

    // set hess to the the Hessian of the likelihood function
    // at pred=meas=obs. The Hessian is the quadratic term in the Taylor
    // expansion of calc_nosystematics(obs, obs), considered as a function of
    // obs (without the factor 1/2). Rows and columns corresponding to
    // unclaimed observables will be zero.
    void get_hessian(Matrix& hess, const ObservableVector& obs);
};


}

#endif

